unit GR32_SeamCarving_Lite;

(* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1 or LGPL 2.1 with linking exception
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Alternatively, the contents of this file may be used under the terms of the
 * Free Pascal modified version of the GNU Lesser General Public License
 * Version 2.1 (the "FPC modified LGPL License"), in which case the provisions
 * of this license are applicable instead of those above.
 * Please see the file LICENSE.txt for additional information concerning this
 * license.
 *
 * AddOn Code for the library Graphics32
 *
 * The initial developer of this AddOn Code is
 * Christian-W. Budde
 *
 * NOTE: This is the lite version. A more complex unit with mask is available
 * from the initial developer!
 *
 * Portions created by the Initial Developer are Copyright (C) 2000-2012
 * the Initial Developer. All Rights Reserved.
 *
 * ***** END LICENSE BLOCK ***** *)

interface

{$I GR32.inc}

uses
  Classes, SysUtils, GR32, GR32_System, GR32_OrdinalMaps;

type
  TWeightCalculation = (wcGradientMagnitude);

  TSeamPath = array of Integer;
  TSeamPaths = array of TSeamPath;

procedure SeamCarving(Source, Destination: TBitmap32;
  MaxSeamsAtOnce: Integer = -1); overload;

implementation

uses
  Math, GR32_LowLevel, GR32_Blend, GR32_Math, GR32_Filters;

procedure QuickSortInteger(Data: PIntegerArray; L, R: Integer);
var
  I, J: Integer;
  P, T: Integer;
begin
 repeat
  I := L;
  J := R;
  P := Data[(L + R) shr 1];
  repeat
    while Data[I] < P do Inc(I);
    while Data[J] > P do Dec(J);
     if I <= J then
      begin
       T := Data[I];
       Data[I] := Data[J];
       Data[J] := T;
       Inc(I);
       Dec(J);
      end;
    until I > J;
   if L < J then QuickSortInteger(Data, L, J);
   L := I;
  until I >= R;
end;

procedure CalculateWeightMap(Bitmap: TBitmap32; WeightMap: TFloatMap);
var
  X, Y: Integer;
  Max, MaxScale: TFloat;
  C: array [0..8] of Byte;
  Value: array [0..1] of Cardinal;
  Data: PFloatArray;

  function CalculateMagnitude(A, B: Integer): Single;
  begin
    Result := FastSqrt(A * A + B * B);
    Assert(Result >= 0);
    if Result > Max then
      Max := Result;
  end;

begin
  if (Bitmap.Width = 0) or (Bitmap.Height = 0) then
    Exit;

  WeightMap.SetSize(Bitmap.Width, Bitmap.Height);

  Max := 0;
  C[1] := Intensity(Bitmap.Pixel[0, 0]);
  C[2] := Intensity(Bitmap.Pixel[1, 0]);
  C[4] := Intensity(Bitmap.Pixel[0, 1]);
  C[5] := Intensity(Bitmap.Pixel[1, 1]);

  Value[0] := (3 * (C[2] - C[1]) + C[5] - C[4]);
  Value[1] := (3 * (C[4] - C[1]) + C[5] - C[2]);

  WeightMap[0, 0] := CalculateMagnitude(Value[0], Value[1]);

  for X := 1 to Bitmap.Width - 2 do
  begin
    C[0] := C[1];
    C[1] := C[2];
    C[2] := Intensity(Bitmap.Pixel[X + 1, 0]);

    C[3] := C[4];
    C[4] := C[5];
    C[5] := Intensity(Bitmap.Pixel[X + 1, 1]);

    Value[0] := (3 * (C[2] - C[0]) + C[5] - C[3]);
    Value[1] := (C[3] + 2 * C[4] + C[5] - (C[0] + 2 * C[1] + C[2]));

    WeightMap[X, 0] := CalculateMagnitude(Value[0], Value[1]);
  end;

  Value[0] := (3 * (C[2] - C[1]) + C[5] - C[4]);
  Value[1] := (3 * (C[5] - C[2]) + C[4] - C[1]);

  WeightMap[Bitmap.Width - 1, 0] := CalculateMagnitude(Value[0], Value[1]);

  for Y := 1 to Bitmap.Height - 2 do
  begin
    C[0] := Intensity(Bitmap.Pixel[0, Y - 1]);
    C[1] := C[0];
    C[2] := Intensity(Bitmap.Pixel[1, Y - 1]);

    C[3] := Intensity(Bitmap.Pixel[0, Y]);
    C[4] := C[3];
    C[5] := Intensity(Bitmap.Pixel[1, Y]);

    C[6] := Intensity(Bitmap.Pixel[0, Y + 1]);
    C[7] := C[6];
    C[8] := Intensity(Bitmap.Pixel[1, Y + 1]);

    Value[0] := (C[2] + 2 * C[5] + C[8] - (C[0] + 2 * C[3] + C[6]));
    Value[1] := (3 * C[6] + C[8] - (3 * C[0] + C[2]));

    WeightMap[0, Y] := CalculateMagnitude(Value[0], Value[1]);

    for X := 1 to Bitmap.Width - 2 do
    begin
      C[0] := C[1];
      C[1] := C[2];
      C[2] := Intensity(Bitmap.Pixel[X + 1, Y - 1]);

      C[3] := C[4];
      C[4] := C[5];
      C[5] := Intensity(Bitmap.Pixel[X + 1, Y    ]);

      C[6] := C[7];
      C[7] := C[8];
      C[8] := Intensity(Bitmap.Pixel[X + 1, Y + 1]);

      Value[0] := (C[2] + 2 * C[5] + C[8] - (C[0] + 2 * C[3] + C[6]));
      Value[1] := (C[6] + 2 * C[7] + C[8] - (C[0] + 2 * C[1] + C[2]));

      WeightMap[X, Y] := CalculateMagnitude(Value[0], Value[1]);
    end;

    Value[0] := (C[2] + 2 * C[5] + C[8] - (C[1] + 2 * C[4] + C[7]));
    Value[1] := (3 * (C[8] - C[2]) + C[7] - C[1]);

    WeightMap[Bitmap.Width - 1, Y] := CalculateMagnitude(Value[0], Value[1]);
  end;

  C[1] := Intensity(Bitmap.Pixel[0, Bitmap.Height - 2]);
  C[2] := Intensity(Bitmap.Pixel[1, Bitmap.Height - 2]);
  C[4] := Intensity(Bitmap.Pixel[0, Bitmap.Height - 1]);
  C[5] := Intensity(Bitmap.Pixel[1, Bitmap.Height - 1]);

  Value[0] := (3 * (C[5] - C[4]) + C[2] - C[1]);
  Value[1] := (3 * (C[4] - C[1]) + C[5] - C[2]);

  WeightMap[0, Bitmap.Height - 1] := CalculateMagnitude(Value[0], Value[1]);

  for X := 1 to Bitmap.Width - 2 do
  begin
    C[0] := C[1];
    C[1] := C[2];
    C[2] := Intensity(Bitmap.Pixel[X + 1, Bitmap.Height - 2]);

    C[3] := C[4];
    C[4] := C[5];
    C[5] := Intensity(Bitmap.Pixel[X + 1, Bitmap.Height - 1]);

    Value[0] := (3 * (C[5] - C[3]) + C[2] - C[0]);
    Value[1] := (C[3] + 2 * C[4] + C[5] - (C[0] + 2 * C[1] + C[2]));

    WeightMap[X, (Bitmap.Height - 1)] := CalculateMagnitude(Value[0], Value[1]);
  end;

  Value[0] := (3 * (C[5] - C[4]) + C[2] - C[1]);
  Value[1] := (3 * (C[5] - C[2]) + C[4] - C[1]);

  WeightMap[Bitmap.Width - 1, Bitmap.Height - 1] := CalculateMagnitude(Value[0], Value[1]);

  MaxScale := 1 / Max;
  Data := WeightMap.Bits;
  for X := 0 to Bitmap.Width * Bitmap.Height - 1 do
    Data^[X] := Data^[X] * MaxScale;
end;

function FindMinimum(Data: PFloatArray; Count: Cardinal): Cardinal;
var
  Index: Cardinal;
  Value: TFloat;
begin
  Assert(Count > 0);
  Result := 0;
  Value := Data^[Result];
  for Index := 1 to Count - 1 do
    if Data^[Index] < Value then
    begin
      Result := Index;
      Value := Data^[Result];
    end;
end;

function SeamPathsValid(var SeamPaths: TSeamPaths): Boolean;
var
  Index, Y: Integer;
  Pixels: array of Cardinal;
begin
  Result := True;
  if Length(SeamPaths) <= 0 then
    Exit;

  SetLength(Pixels, Length(SeamPaths));
  for Y := 0 to Length(SeamPaths[0]) - 1 do
  begin
    for Index := 0 to Length(SeamPaths) - 1 do
      Pixels[Index] := SeamPaths[Index, Y];
    QuickSortInteger(PIntegerArray(@Pixels[0]), 0, Length(SeamPaths) - 1);
    for Index := 1 to Length(SeamPaths) - 1 do
      if Pixels[Index] = Pixels[Index - 1] then
      begin
        Result := False;
        Exit;
      end;
  end;
end;

type
  PFloat2Array = ^TFloat2Array;
  TFloat2Array = array [0..1] of TFloat;

function CalculateVerticalSeams(WeightMap: TFloatMap; var SeamPaths: TSeamPaths;
  Count: Cardinal): Cardinal;
var
  Index, X, Y, W, H: Integer;
  Value, Test: TFloat;
  SumMap: PFloatArray;
  SumMapPtr: array [0..1] of PFloatArray;
  Dir: Integer;
  Penalty: TFloat;
  TempWeights: TFloatMap;
begin
  if Count <= 0 then
    Exit;

  W := WeightMap.Width;
  H := WeightMap.Height;
  if (W = 0) or (H = 0) then
    Exit;

  SetLength(SeamPaths, Count);

  Penalty := 1E23;

  TempWeights := TFloatMap.Create;
  TempWeights.SetSize(W, H);
  try
    Move(WeightMap.Bits^, TempWeights.Bits^, W * H * SizeOf(TFloat));
    GetMem(SumMap, 2 * W * SizeOf(TFloat));
    try
      for Index := 0 to Count - 1 do
      begin
        SumMapPtr[0] := @SumMap^[0];
        SumMapPtr[1] := @SumMap^[W];
        Move(TempWeights.Bits^, SumMap^[0], W * SizeOf(TFloat));

        for Y := 1 to H - 1 do
        begin
          Swap(Pointer(SumMapPtr[0]), Pointer(SumMapPtr[1]));

          if TempWeights[1, Y - 1] < TempWeights[0, Y - 1] then
            SumMapPtr[0]^[0] := PFloat2Array(SumMapPtr[1])^[1] + TempWeights[0, Y]
          else
            SumMapPtr[0]^[0] := SumMapPtr[1]^[0] + TempWeights[0, Y];

          for X := 1 to W - 2 do
          begin
            Dir := 0;
            Value := TempWeights[X, Y - 1];
            if TempWeights[X - 1, Y - 1] < Value then
            begin
              Value := TempWeights[X - 1, Y - 1];
              Dir := -1;
            end;
            if TempWeights[X + 1, Y - 1] < Value then
              Dir := +1;
            SumMapPtr[0]^[X] := SumMapPtr[1]^[X + Dir] + TempWeights[X, Y];
          end;

          if TempWeights[W - 2, Y - 1] < TempWeights[W - 1, Y - 1] then
            SumMapPtr[0]^[W - 1] := SumMapPtr[1]^[W - 2] + TempWeights[W - 1, Y]
          else
            SumMapPtr[0]^[W - 1] := SumMapPtr[1]^[W - 1] + TempWeights[W - 1, Y];
        end;

        Y := H - 1;
        SetLength(SeamPaths[Index], H);
        X := FindMinimum(@SumMapPtr[0]^[0], W);
        if SumMapPtr[0]^[X] >= Penalty then
        begin
          SetLength(SeamPaths, Index);
          Result := Index;
          Assert(SeamPathsValid(SeamPaths));
          Exit;
        end;

        SeamPaths[Index][Y] := X;
        TempWeights[X, Y] := Penalty;

        for Y := H - 2 downto 0 do
        begin
          Dir := 0;
          Value := TempWeights[X, Y];
          if (X > 0) and (TempWeights[X - 1, Y] < Value) then
          begin
            Dir := - 1;
            Value := TempWeights[X - 1, Y];
          end;
          if (X < W - 1) and (TempWeights[X + 1, Y] < Value) then
            Dir := 1;

          X := X + Dir;
          SeamPaths[Index][Y] := X;
          TempWeights[X, Y] := Penalty;
        end;
      end;
    finally
      FreeMem(SumMap);
    end;
  finally
    FreeAndNil(TempWeights);
  end;

  Assert(SeamPathsValid(SeamPaths));
  Result := Count;
end;

function CalculateHorizontalSeams(WeightMap: TFloatMap; var SeamPaths: TSeamPaths;
  Count: Cardinal): Cardinal;
var
  Index, X, Y, W, H: Integer;
  Value, Test: TFloat;
  SumMap: PFloatArray;
  SumMapPtr: array [0..1] of PFloatArray;
  Dir: Integer;
  Penalty: TFloat;
  TempWeights: TFloatMap;
begin
  if Count <= 0 then
    Exit;

  SetLength(SeamPaths, Count);

  W := WeightMap.Width;
  H := WeightMap.Height;

  Penalty := 1E23;

  TempWeights := TFloatMap.Create;
  TempWeights.SetSize(W, H);
  try
    Move(WeightMap.Bits^, TempWeights.Bits^, W * H * SizeOf(TFloat));
    GetMem(SumMap, 2 * H * SizeOf(TFloat));
    try
      for Index := 0 to Count - 1 do
      begin
        SumMapPtr[0] := @SumMap^[0];
        SumMapPtr[1] := @SumMap^[H];
        for Y := 0 to H - 1 do
          SumMapPtr[0]^[Y] := WeightMap.Value[0, Y];

        for X := 1 to W - 1 do
        begin
          Swap(Pointer(SumMapPtr[0]), Pointer(SumMapPtr[1]));

          if TempWeights[X - 1, 1] < TempWeights[X - 1, 0] then
            SumMapPtr[0]^[0] := PFloat2Array(SumMapPtr[1])^[1] + TempWeights[X, 0]
          else
            SumMapPtr[0]^[0] := SumMapPtr[1]^[0] + TempWeights[X, 0];

          for Y := 1 to H - 2 do
          begin
            Dir := 0;
            Value := TempWeights[X - 1, Y];
            if TempWeights[X - 1, Y - 1] < Value then
            begin
              Value := TempWeights[X - 1, Y - 1];
              Dir := -1;
            end;
            if TempWeights[X - 1, Y + 1] < Value then
              Dir := +1;
            SumMapPtr[0]^[Y] := SumMapPtr[1]^[Y + Dir] + TempWeights[X, Y];
          end;

          if TempWeights[X - 1, H - 2] < TempWeights[X - 1, H - 1] then
            SumMapPtr[0]^[H - 1] := SumMapPtr[1]^[H - 2] + TempWeights[X, H - 1]
          else
            SumMapPtr[0]^[H - 1] := SumMapPtr[1]^[H - 1] + TempWeights[X, H - 1];
        end;

        X := W - 1;
        SetLength(SeamPaths[Index], W);
        Y := FindMinimum(@SumMapPtr[0]^[0], H);
        if SumMapPtr[0]^[Y] >= Penalty then
        begin
          SetLength(SeamPaths, Index);
          Result := Index;
          Assert(SeamPathsValid(SeamPaths));
          Exit;
        end;

        SeamPaths[Index][X] := Y;
        TempWeights[X, Y] := Penalty;

        for X := W - 2 downto 0 do
        begin
          Dir := 0;
          Value := TempWeights[X, Y];
          if (Y > 0) and (TempWeights[X, Y - 1] < Value) then
          begin
            Dir := - 1;
            Value := TempWeights[X, Y - 1];
          end;
          if (Y < H - 1) and (TempWeights[X, Y + 1] < Value) then
            Dir := 1;

          Y := Y + Dir;
          SeamPaths[Index][X] := Y;
          TempWeights[X, Y] := Penalty;
        end;
      end;
    finally
      FreeMem(SumMap);
    end;
  finally
    FreeAndNil(TempWeights);
  end;

  Assert(SeamPathsValid(SeamPaths));
  Result := Count;
end;

procedure ShrinkWidth(Source, Destination: TBitmap32; SeamPaths: TSeamPaths); overload;
var
  Index, X, Y, W, H, Last: Integer;
  ScanLine: array [0..1] of PColor32Array;
  Pixels: array of Integer;
begin
  W := Source.Width;
  H := Source.Height;
  Destination.SetSize(W - Length(SeamPaths), H);

  SetLength(Pixels, Length(SeamPaths));

  for Y := 0 to H - 1 do
  begin
    for Index := 0 to Length(SeamPaths) - 1 do
      Pixels[Index] := SeamPaths[Index, Y];
    QuickSortInteger(PIntegerArray(@Pixels[0]), 0, Length(SeamPaths) - 1);

    ScanLine[0] := Source.ScanLine[Y];
    ScanLine[1] := Destination.ScanLine[Y];

    X := Pixels[0];
    Move(ScanLine[0]^[0], ScanLine[1]^[0], X * SizeOf(TColor32));

    for Index := 1 to Length(SeamPaths) - 1 do
    begin
      Last := X + 1;
      X := Pixels[Index];

      Move(ScanLine[0]^[Last], ScanLine[1]^[Last - Index],
        (X - Last) * SizeOf(TColor32));
    end;

    Last := X + 1;
    Move(ScanLine[0]^[Last], ScanLine[1]^[Last - Length(SeamPaths)],
      (W - Last) * SizeOf(TColor32));
  end;
end;

procedure ShrinkWidth(Bitmap: TBitmap32; SeamPaths: TSeamPaths); overload;
var
  Temp: TBitmap32;
begin
  Temp := TBitmap32.Create;
  try
    ShrinkWidth(Bitmap, Temp, SeamPaths);
    Bitmap.Assign(Temp);
  finally
    Temp.Free;
  end;
end;

procedure ShrinkHeight(Source, Destination: TBitmap32; SeamPaths: TSeamPaths); overload;
var
  Index, X, Y, W, H, Cur, Last, Count: Integer;
  Pixels: array of Integer;
begin
  Count := Length(SeamPaths);
  W := Source.Width;
  H := Source.Height;
  Destination.SetSize(W, H - Count);
  SetLength(Pixels, Count);

  for X := 0 to W - 1 do
  begin
    for Index := 0 to Count - 1 do
      Pixels[Index] := SeamPaths[Index, X];
    QuickSortInteger(PIntegerArray(@Pixels[0]), 0, Count - 1);

    Cur := Pixels[0];
    for Y := 0 to Cur - 1 do
      Destination.Pixel[X, Y] := Source.Pixel[X, Y];

    for Index := 1 to Count - 1 do
    begin
      Last := Cur + 1;
      Cur := Pixels[Index];

      for Y := Last to Cur - 1 do
        Destination.Pixel[X, Y - Index] := Source.Pixel[X, Y];
    end;

    Last := Cur + 1;
    for Y := Last to H - 1 do
      Destination.Pixel[X, Y - Count] := Source.Pixel[X, Y];
  end;
end;

procedure ShrinkHeight(Bitmap: TBitmap32; SeamPaths: TSeamPaths); overload;
var
  Temp: TBitmap32;
begin
  Temp := TBitmap32.Create;
  try
    ShrinkHeight(Bitmap, Temp, SeamPaths);
    Bitmap.Assign(Temp);
  finally
    Temp.Free;
  end;
end;

procedure ExpandWidth(Source, Destination: TBitmap32; SeamPaths: TSeamPaths); overload;
var
  Index, X, Y, W, H, Last, Count: Integer;
  ScanLine: array [0..1] of PColor32Array;
  Pixels: array of Integer;
begin
  Count := Length(SeamPaths);
  W := Source.Width;
  H := Source.Height;
  Destination.SetSize(W + Count, H);
  SetLength(Pixels, Count);

  for Y := 0 to H - 1 do
  begin
    for Index := 0 to Count - 1 do
      Pixels[Index] := SeamPaths[Index, Y];
    QuickSortInteger(PIntegerArray(@Pixels[0]), 0, Count - 1);

    ScanLine[0] := Source.ScanLine[Y];
    ScanLine[1] := Destination.ScanLine[Y];

    X := Pixels[0];
    Move(ScanLine[0]^[0], ScanLine[1]^[0], X * SizeOf(TColor32));

    for Index := 1 to Count - 1 do
    begin
      Last := X;
      X := Pixels[Index];

      if Last - 1 < 0 then
        ScanLine[1]^[Last + Index - 1] := ScanLine[0]^[Last]
      else
        ScanLine[1]^[Last + Index - 1] := CombineReg(
          ScanLine[0]^[Last - 1], ScanLine[0]^[Last], $7F);

      Move(ScanLine[0]^[Last], ScanLine[1]^[Last + Index],
        (X - Last) * SizeOf(TColor32));
    end;

    Last := X;

    if Last - 1 < 0 then
      ScanLine[1]^[Last + Count - 1] := ScanLine[0]^[Last]
    else
      ScanLine[1]^[Last + Count - 1] := CombineReg(
        ScanLine[0]^[Last - 1], ScanLine[0]^[Last], $7F);

    Move(ScanLine[0]^[Last], ScanLine[1]^[Last + Count],
      (W - Last) * SizeOf(TColor32));
  end;
end;

procedure ExpandWidth(Bitmap: TBitmap32; SeamPaths: TSeamPaths); overload;
var
  Temp: TBitmap32;
begin
  Temp := TBitmap32.Create;
  try
    ExpandWidth(Bitmap, Temp, SeamPaths);
    Bitmap.Assign(Temp);
  finally
    Temp.Free;
  end;
end;

procedure ExpandHeight(Source, Destination: TBitmap32; SeamPaths: TSeamPaths); overload;
var
  Index, X, Y, W, H, Last, Cur, Count: Integer;
  Pixels: array of Integer;
begin
  Count := Length(SeamPaths);
  W := Source.Width;
  H := Source.Height;
  Destination.SetSize(W, H + Count);
  SetLength(Pixels, Count);

  for X := 0 to W - 1 do
  begin
    for Index := 0 to Count - 1 do
      Pixels[Index] := SeamPaths[Index, X];
    QuickSortInteger(PIntegerArray(@Pixels[0]), 0, Count - 1);

    Cur := Pixels[0];
    for Y := 0 to Cur - 1 do
      Destination.Pixel[X, Y] := Source.Pixel[X, Y];

    for Index := 1 to Count - 1 do
    begin
      Last := Cur;
      Cur := Pixels[Index];

      if Last - 1 < 0 then
        Destination.Pixel[X, Last + Index - 1] := Source.Pixel[X, Last]
      else
        Destination.Pixel[X, Last + Index - 1] := CombineReg(
          Source.Pixel[X, Last - 1], Source.Pixel[X, Last], $7F);

      for Y := Last to Cur - 1 do
        Destination.Pixel[X, Y + Index] := Source.Pixel[X, Y];
    end;

    Last := Cur;

    if Last - 1 < 0 then
      Destination.Pixel[X, Last + Count - 1] := Source.Pixel[X, Last]
    else
      Destination.Pixel[X, Last + Count - 1] := CombineReg(
        Source.Pixel[X, Last - 1], Source.Pixel[X, Last], $7F);

    for Y := Last to H - 1 do
      Destination.Pixel[X, Y + Count] := Source.Pixel[X, Y];
  end;
end;

procedure ExpandHeight(Bitmap: TBitmap32; SeamPaths: TSeamPaths); overload;
var
  Temp: TBitmap32;
begin
  Temp := TBitmap32.Create;
  try
    ExpandHeight(Bitmap, Temp, SeamPaths);
    Bitmap.Assign(Temp);
  finally
    Temp.Free;
  end;
end;

procedure ChangeWidth(Source, Destination: TBitmap32;
  MaxSeamsAtOnce: Integer = -1); overload;
var
  WeightMap: TFloatMap;
  Paths: TSeamPaths;
  IsShrink: Boolean;
  Delta, Seams: Integer;
  Temp: TBitmap32;
begin
  if (Source.Width = Destination.Width) then
  begin
    Move(Source.Bits^, Destination.Bits^,
      Source.Width * Min(Source.Height, Destination.Height) * SizeOf(TColor32));
    Exit;
  end;

  Delta := Abs(Source.Width - Destination.Width);
  if MaxSeamsAtOnce < 1 then
    MaxSeamsAtOnce := Delta;

  WeightMap := TFloatMap.Create;
  try
    CalculateWeightMap(Source, WeightMap);

    IsShrink := Destination.Width < Source.Width;
    Seams := CalculateVerticalSeams(WeightMap, Paths,
      Min(Delta, MaxSeamsAtOnce));
    Assert(Seams <= Delta);

    if Delta = Seams then
    begin
      if IsShrink then
        ShrinkWidth(Source, Destination, Paths)
      else
        ExpandWidth(Source, Destination, Paths);
      Exit;
    end;

    Temp := TBitmap32.Create;
    try
      if IsShrink then
        ShrinkWidth(Source, Temp, Paths)
      else
        ExpandWidth(Source, Temp, Paths);
      Delta := Delta - Seams;
      CalculateWeightMap(Temp, WeightMap);
      Seams := CalculateVerticalSeams(WeightMap, Paths,
        Min(Delta, MaxSeamsAtOnce));
      Assert(Seams <= Delta);

      if Seams < Delta then
      repeat
        if IsShrink then
          ShrinkWidth(Temp, Paths)
        else
          ExpandWidth(Temp, Paths);
        Delta := Delta - Seams;
        CalculateWeightMap(Temp, WeightMap);
        Seams := CalculateVerticalSeams(WeightMap, Paths,
          Min(Delta, MaxSeamsAtOnce));
        Assert(Seams <= Delta);
      until Delta = Seams;

      if IsShrink then
        ShrinkWidth(Temp, Destination, Paths)
      else
        ExpandWidth(Temp, Destination, Paths);
    finally
      Temp.Free;
    end;
  finally
    WeightMap.Free;
  end;
end;

procedure ChangeHeight(Source, Destination: TBitmap32;
  MaxSeamsAtOnce: Integer = -1); overload;
var
  WeightMap: TFloatMap;
  Paths: TSeamPaths;
  IsShrink: Boolean;
  Index: Integer;
  Delta, Seams: Integer;
  Temp: TBitmap32;
begin
  if (Source.Height = Destination.Height) then
  begin
    for Index := 0 to Source.Height - 1 do
    begin
      Move(Source.Bits^, Destination.Bits^,
        Min(Source.Width, Destination.Width) * SizeOf(TColor32));
    end;
    Exit;
  end;

  Delta := Abs(Source.Height - Destination.Height);
  if MaxSeamsAtOnce < 1 then
    MaxSeamsAtOnce := Delta;

  WeightMap := TFloatMap.Create;
  try
    CalculateWeightMap(Source, WeightMap);

    IsShrink := Destination.Height < Source.Height;
    Seams := CalculateHorizontalSeams(WeightMap, Paths,
      Min(Delta, MaxSeamsAtOnce));
    Assert(Seams <= Delta);

    if Delta = Seams then
    begin
      if IsShrink then
        ShrinkHeight(Source, Destination, Paths)
      else
        ExpandHeight(Source, Destination, Paths);
      Exit;
    end;

    Temp := TBitmap32.Create;
    try
      if IsShrink then
        ShrinkHeight(Source, Temp, Paths)
      else
        ExpandHeight(Source, Temp, Paths);
      Delta := Delta - Seams;
      CalculateWeightMap(Temp, WeightMap);
      Seams := CalculateHorizontalSeams(WeightMap, Paths,
        Min(Delta, MaxSeamsAtOnce));
      Assert(Seams <= Delta);

      if Seams < Delta then
      repeat
        if IsShrink then
          ShrinkHeight(Temp, Paths)
        else
          ExpandHeight(Temp, Paths);
        Delta := Delta - Seams;
        CalculateWeightMap(Temp, WeightMap);
        Seams := CalculateHorizontalSeams(WeightMap, Paths,
          Min(Delta, MaxSeamsAtOnce));
        Assert(Seams <= Delta);
      until Delta = Seams;

      if IsShrink then
        ShrinkHeight(Temp, Destination, Paths)
      else
        ExpandHeight(Temp, Destination, Paths);
    finally
      Temp.Free;
    end;
  finally
    WeightMap.Free;
  end;
end;

procedure SeamCarving(Source: TBitmap32; Destination: TBitmap32;
  MaxSeamsAtOnce: Integer = -1);
var
  Temp: TBitmap32;
begin
  if (Source.Width <> Destination.Width) then
  begin
    if (Source.Height <> Destination.Height) then
    begin
      Temp := TBitmap32.Create;
      try
        Temp.SetSize(Source.Width, Destination.Height);
        ChangeHeight(Source, Temp, MaxSeamsAtOnce);
        ChangeWidth(Temp, Destination, MaxSeamsAtOnce);
      finally
        Temp.Free;
      end;
    end
    else
      ChangeWidth(Source, Destination, MaxSeamsAtOnce)
  end else
  if (Source.Height <> Destination.Height) then
    ChangeHeight(Source, Destination, MaxSeamsAtOnce)
  else
  begin
    Move(Source.Bits^, Destination.Bits^,
      Source.Width * Source.Height * SizeOf(TColor32));
  end;
end;

end.
