unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtDlgs, Menus, GR32, GR32_Resamplers, GR32_SeamCarving_Lite;

type
  TRescaleUpdateThread = class(TThread)
  private
    FTemp: TBitmap32;
    procedure PerformRescale;
    procedure Update;
  protected
    procedure Execute; override;
  end;

  TSeamCarveUpdateThread = class(TThread)
  private
    FCopy : TBitmap32;
    FTemp : TBitmap32;
    procedure PerformSeamCarving;
    procedure Update;
    procedure CopyRescaleBitmap;
  protected
    procedure Execute; override;
  end;

  TFmSeamCarver = class(TForm)
    OpenPictureDialog: TOpenPictureDialog;
    PopupMenu: TPopupMenu;
    MiShowCropped: TMenuItem;
    N1: TMenuItem;
    MiShrinkOnly: TMenuItem;
    MiEnlargeOnly: TMenuItem;
    MiAuto: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormClick(Sender: TObject);
    procedure MiShowCroppedClick(Sender: TObject);
    procedure MiShrinkOnlyClick(Sender: TObject);
    procedure MiEnlargeOnlyClick(Sender: TObject);
    procedure MiAutoClick(Sender: TObject);
  private
    FFileName: TFileName;
    FBitmap: TBitmap32;
    FRescale: TBitmap32;
    FBuffer: TBitmap32;
    FRescaleRatio: TFloat;

    FRescaleUpdateThread: TRescaleUpdateThread;
    FSeamCarveUpdateThread: TSeamCarveUpdateThread;
    procedure SizeChanged;
  public
    NeedsRescale: Boolean;
    NeedsSeamCarving: Boolean;

    property RescaleRatio: TFloat read FRescaleRatio;
    property OriginalBitmap: TBitmap32 read FBitmap;
  end;

var
  FmSeamCarver: TFmSeamCarver;

implementation

uses
  Math, JPEG, GR32_LowLevel;

{$R *.dfm}

{ TRescaleUpdateThread }

procedure TRescaleUpdateThread.Execute;
begin
  FTemp := nil;
  while not Terminated do
  begin
    PerformRescale;
    Synchronize(Update);

    if not FmSeamCarver.NeedsRescale then
      Suspended := True;
  end;
end;

procedure TRescaleUpdateThread.PerformRescale;
var
  DestRect: TRect;
begin
  with FmSeamCarver do
  begin
    NeedsRescale := False;

    Assert(FTemp = nil);
    FTemp := TBitmap32.Create;
    DestRect := Rect(0, 0, Round(RescaleRatio * OriginalBitmap.Width),
      Round(RescaleRatio * OriginalBitmap.Height));
    FTemp.SetSize(DestRect.Right - DestRect.Left, DestRect.Bottom - DestRect.Top);
    FTemp.Draw(DestRect, OriginalBitmap.BoundsRect, FBitmap);
  end;
end;

procedure TRescaleUpdateThread.Update;
begin
  with FmSeamCarver do
  begin
    if MiShowCropped.Checked then
    begin
      FBuffer.Assign(FTemp);
      Caption := FFileName + ' - Cropped! Please wait while performing seam carving...';
      Invalidate;
    end;
    Swap(Pointer(FRescale), Pointer(FTemp));
    FreeAndNil(FTemp);

    NeedsSeamCarving := True;
    FSeamCarveUpdateThread.Suspended := False;
  end;
end;


{ TSeamCarveUpdateThread }

procedure TSeamCarveUpdateThread.Execute;
begin
  FTemp := nil;
  FCopy := TBitmap32.Create;
  try
    while not Terminated do
    begin
      Synchronize(CopyRescaleBitmap);
      if Terminated then Exit;
      PerformSeamCarving;
      Synchronize(Update);
      if Terminated then Exit;
      if not FmSeamCarver.NeedsSeamCarving then
        Suspended := True;
    end;
  finally
    FCopy.Free;
  end;
end;

procedure TSeamCarveUpdateThread.CopyRescaleBitmap;
begin
  FCopy.Assign(FmSeamCarver.FRescale);
end;


procedure TSeamCarveUpdateThread.PerformSeamCarving;
begin
  with FmSeamCarver do
  begin
    NeedsSeamCarving := False;

    FTemp := TBitmap32.Create;
    FTemp.SetSize(ClientWidth, ClientHeight);
    SeamCarving(FCopy, FTemp);
  end;
end;

procedure TSeamCarveUpdateThread.Update;
begin
  with FmSeamCarver do
  begin
    if not NeedsSeamCarving then
    begin
      Swap(Pointer(FBuffer), Pointer(FTemp));
      Caption := FFileName + ' - Seam Carved!';
      Invalidate;
    end;
    FreeAndNil(FTemp);
  end;
end;


{ TFmSeamCarver }

procedure TFmSeamCarver.FormCreate(Sender: TObject);
var
  KernelResampler: TKernelResampler;
begin
  ControlStyle := ControlStyle + [csOpaque];

  KernelResampler := TKernelResampler.Create;
  KernelResampler.Kernel := TLanczosKernel.Create;

  FFileName := 'Sample.jpg';

  FBitmap := TBitmap32.Create;
  FBitmap.LoadFromFile(FFileName);
  FBitmap.Resampler := KernelResampler;

  FRescale := TBitmap32.Create;
  FRescale.Assign(FBitmap);

  FBuffer := TBitmap32.Create;
  FBuffer.Assign(FBitmap);

  FRescaleUpdateThread := TRescaleUpdateThread.Create(True);
  FSeamCarveUpdateThread := TSeamCarveUpdateThread.Create(True);

  OnResize := nil;

  ClientWidth := FBitmap.Width;
  ClientHeight := FBitmap.Height;

  NeedsRescale := False;
  NeedsSeamCarving := False;
end;

procedure TFmSeamCarver.FormDestroy(Sender: TObject);
begin
  FRescaleUpdateThread.Terminate;
  if FRescaleUpdateThread.Suspended then
    FRescaleUpdateThread.Suspended := False;
  FRescaleUpdateThread.WaitFor;

  FSeamCarveUpdateThread.Terminate;
  if FSeamCarveUpdateThread.Suspended then
    FSeamCarveUpdateThread.Suspended := False;
  FSeamCarveUpdateThread.WaitFor;

  FreeAndNil(FRescaleUpdateThread);
  FreeAndNil(FSeamCarveUpdateThread);
  FreeAndNil(FBuffer);
  FreeAndNil(FRescale);
  FreeAndNil(FBitmap);
end;

procedure TFmSeamCarver.FormShow(Sender: TObject);
begin
  OnResize := FormResize;
end;

procedure TFmSeamCarver.MiAutoClick(Sender: TObject);
begin
  if not MiAuto.Checked then
  begin
    MiAuto.Checked := True;
    SizeChanged;
  end;
end;

procedure TFmSeamCarver.MiEnlargeOnlyClick(Sender: TObject);
begin
  if not MiEnlargeOnly.Checked then
  begin
    MiEnlargeOnly.Checked := True;
    SizeChanged;
  end;
end;

procedure TFmSeamCarver.MiShowCroppedClick(Sender: TObject);
begin
  MiShowCropped.Checked := not MiShowCropped.Checked;
end;

procedure TFmSeamCarver.MiShrinkOnlyClick(Sender: TObject);
begin
  if not MiShrinkOnly.Checked then
  begin
    MiShrinkOnly.Checked := True;
    SizeChanged;
  end;
end;

procedure TFmSeamCarver.FormResize(Sender: TObject);
begin
  SizeChanged;
end;

procedure TFmSeamCarver.SizeChanged;
begin
  if MiShrinkOnly.Checked then
    FRescaleRatio := Max(ClientHeight / FBitmap.Height,
      ClientWidth / FBitmap.Width)
  else
  if MiEnlargeOnly.Checked then
    FRescaleRatio := Min(ClientHeight / FBitmap.Height,
      ClientWidth / FBitmap.Width)
  else
    FRescaleRatio := 0.5 * (ClientHeight / FBitmap.Height +
      ClientWidth / FBitmap.Width);

  if FRescaleRatio = 1 then
  begin
    FBuffer.Assign(FBitmap);
    FmSeamCarver.Caption := FmSeamCarver.FFileName;
    Exit;
  end;

  NeedsRescale := True;
  FRescaleUpdateThread.Suspended := False;
end;

procedure TFmSeamCarver.FormPaint(Sender: TObject);
begin
  FBuffer.DrawTo(Canvas.Handle, 0, 0);
end;

procedure TFmSeamCarver.FormClick(Sender: TObject);
begin
  if OpenPictureDialog.Execute then
  begin
    FBitmap.LoadFromFile(OpenPictureDialog.FileName);
    OnResize := nil;
    ClientWidth := FBitmap.Width;
    ClientHeight := FBitmap.Height;
    OnResize := FormResize;
    Invalidate;
  end;
end;

end.
